//!Beanshell

/* Imports */
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.Date;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.lang.StringBuilder;

import de.espirit.firstspirit.agency.BrokerAgent;
import de.espirit.firstspirit.agency.StoreAgent;
import de.espirit.firstspirit.agency.OperationAgent;
import de.espirit.firstspirit.agency.SpecialistsBroker;
import de.espirit.firstspirit.agency.StoreElementAgent;

import de.espirit.firstspirit.ui.operations.RequestOperation;

import de.espirit.firstspirit.access.Language;
import de.espirit.firstspirit.access.AccessUtil;
import de.espirit.firstspirit.access.Principal;
import de.espirit.firstspirit.access.project.Group;
import de.espirit.firstspirit.access.ReferenceEntry;

import de.espirit.firstspirit.access.store.Store;
import de.espirit.firstspirit.access.store.IDProvider;
import de.espirit.firstspirit.access.store.IDProvider.UidType;
import de.espirit.firstspirit.access.store.Permission;
import de.espirit.firstspirit.access.store.StoreElement;
import de.espirit.firstspirit.access.store.templatestore.Template;
import de.espirit.firstspirit.access.store.templatestore.TemplateFolder;
import de.espirit.firstspirit.access.store.templatestore.SectionTemplate;
import de.espirit.firstspirit.access.store.templatestore.FormatTemplate;
import de.espirit.firstspirit.access.store.pagestore.Page;
import de.espirit.firstspirit.access.store.pagestore.PageFolder;
import de.espirit.firstspirit.access.store.pagestore.Body;
import de.espirit.firstspirit.access.store.pagestore.Section;
import de.espirit.firstspirit.access.store.sitestore.PageRef;
import de.espirit.firstspirit.access.store.sitestore.PageRefFolder;
import de.espirit.firstspirit.access.store.mediastore.Media;
import de.espirit.firstspirit.access.store.mediastore.MediaFolder;
import de.espirit.firstspirit.access.store.contentstore.Content2;
import de.espirit.firstspirit.access.store.contentstore.ContentFolder;

import de.espirit.firstspirit.access.editor.value.TargetReference;

import de.espirit.firstspirit.forms.FormData;
import de.espirit.firstspirit.forms.FormField;
import de.espirit.firstspirit.forms.FormDataList;

/* 
* This script makes a copy of the current page and places it in the location specified by 'destination'.
* It adds a link on the copy to the original page, and marks the copy as 'archived'
* If this is the first time that the current page has been copied in this way, it adds a link to the original to the destination 'previous versions'
* /

// context = de.espirit.firstspirit.webedit.server.workflow.WebeditScriptContextImpl
// originalPublishDate = de.espirit.firstspirit.webedit.server.forms.StoreElementFormServiceImpl$FormScriptFormField - aka de.espirit.firstspirit.forms.FormField<T>
// destination = de.espirit.firstspirit.webedit.server.forms.StoreElementFormServiceImpl$FormScriptFormField

/* Parameters */
// "destination.get()": de.espirit.firstspirit.access.editor.value.TargetReference

/* Global declarations */

currentPage = context.getElement(); // The current, source page
final Language lang = context.getProject().getMasterLanguage();
final SimpleDateFormat dateFormat = new SimpleDateFormat("dd MM yyyy");
final SimpleDateFormat prettyDateFormat = new SimpleDateFormat("d MMM, yyyy");
Calendar today = Calendar.getInstance();
final static String THIS_IS_OLD = "<b>This is an old version of the page</b>";
final static String THIS_IS_OLD_META_DESC = "[This is an old version of the page]";
final static String CURRENT_GUIDANCE = "Current guidance";
final static String HEADING_PREV_VERSIONS = "Previous versions";
final static String LINK_PREV_VERSIONS = "Previous versions of this page";
final StoreAgent storeAgent = context.requireSpecialist(StoreAgent.TYPE);
final Store pageStore = storeAgent.getStore(Store.Type.PAGESTORE);
final Store siteStore = storeAgent.getStore(Store.Type.SITESTORE);
final Store templateStore = storeAgent.getStore(Store.Type.TEMPLATESTORE);

/* Methods */

// Display a message
showMessage(message) { 
	requestOperation = context.requireSpecialist(OperationAgent.TYPE).getOperation(RequestOperation.TYPE); 
	requestOperation.setKind(RequestOperation.Kind.INFO); 
	requestOperation.perform(message);
}

// create a folder in the page store
PageFolder createPageStoreFolder(uid, String name) {
	PageFolder parentPageFolder = pageStore.getStoreElement(uid,IDProvider.UidType.PAGESTORE);
	
	try {
		parentPageFolder.setLock(true,false);
		newPageFolder = parentPageFolder.createPageFolder(name, true);
		newPageFolder.setDisplayName(lang,name);
		newPageFolder.save("Script - create archive page folder for " + uid, true);
		parentPageFolder.setLock(false,false);
	} catch (Exception ex) {
		String error = "Exception while creating PageStoreFolder under " + pageStoreFolder.getUid() + "\n Exception is : " + ex.getMessage();
		showMessage(error);
	}
	
	return newPageFolder;
}

// create a copy of page into folder
copyCurrentPage(IDProvider folder, Page page, String name) {
	Page newPage;
	try {
		folder.setLock(true,false);
		page.setLock(true,false);
		newPage = AccessUtil.copyStoreElement(page,folder);
		newPage.setLock(true,false);
		newPage.setDisplayName(lang, name);
		newPage.save();
		newPage.setLock(false,false);
		page.setLock(false,false);
		folder.setLock(false,false);
	} catch (Exception ex) {
		String error = "Exception while copying page " + page.getUid() + "\n Exception is : " + ex.getMessage();
		showMessage(error);
	}
	return newPage;
}

// create a section containing information about the archive version dates
addArchivedInformationSections(Page page, Calendar today, PageRef original) {
	page.setLock(true,false);
	
	// mark this page as archived, clear archive location reference
	FormData pageFormData = page.getFormData();
	pageFormData.get(lang,"pt_archived").set(true);
	pageFormData.get(lang,"sc_copy_destination").setToDefault();
	pageFormData.get(lang,"pt_heading_description").set(THIS_IS_OLD_META_DESC);
	Calendar tomorrow = Calendar.getInstance();
	tomorrow = today; // just in case this script is run over midnight...
	tomorrow.add(Calendar.DAY_OF_MONTH,1);
	pageFormData.get(lang,"st_superseded_on").set(tomorrow.getTime());
	page.setFormData(pageFormData);
	
	// if there is no related content link section, create one
	Body relatedContent = page.getBodyByName("pt_related_topics_list");
	Section guidanceSection;
	try {
		if(relatedContent.getChildCount() == 0) {
			SectionTemplate st_related_content = templateStore.getStoreElement("st_related_content",IDProvider.UidType.TEMPLATESTORE);
			guidanceSection = relatedContent.createSection("Current_guidance",st_related_content);
		} else {
			guidanceSection = relatedContent.getFirstChild();
		}
		guidanceSection.setLock(true,false);
		FormData rcFormData = guidanceSection.getFormData();
		FormField linksField = rcFormData.get(lang,"st_related_content_links");
		
		FormDataList dataList = linksField.get();
		producer = dataList.getProducer();
		
		// programmatically creating links is quite a bit of effort
		lt_internal_link = templateStore.getStoreElement("lt_internal_link",IDProvider.UidType.TEMPLATESTORE_LINKTEMPLATE);
		linkFormData = producer.create(lt_internal_link,lang);
		lt_link_type = linkFormData.get(lang,"lt_link_type");
		lt_link_type.set(true);
		lt_title = linkFormData.get(lang,"lt_title");
		lt_title.set(CURRENT_GUIDANCE);
		lt_text = linkFormData.get(lang,"lt_text");
		lt_text.set(CURRENT_GUIDANCE);
		lt_reference = linkFormData.get(lang,"lt_reference");
		// this is how you set an FS_REFERENCE value
		originalRef = TargetReference.TargetReferences.newInstance(lang, original, null);
		lt_reference.set(originalRef);
		
		dataList.add(linkFormData);
		linksField.set(dataList);

		guidanceSection.setFormData(rcFormData);
		guidanceSection.save();
		guidanceSection.setLock(false,false);
	} catch (Exception sectionex) {
		showMessage(sectionex.getMessage());
		context.logError("Exception while creating related links guidance section on page archive.");
		context.logError(sectionex.getMessage());
	}
	
	// check if there is a date published value
	/*if(pageFormData.get(lang,"pt_publish_date") != null && !pageFormData.get(lang,"pt_publish_date").isEmpty()) {
		try {
			Body sectionContent = page.getBodyByName("pt_section_content");
			SectionTemplate formattedTextSectionTemplate = templateStore.getStoreElement("st_formatted_text",IDProvider.UidType.TEMPLATESTORE);
			Section archiveSection = sectionContent.createSection("archived_notes",formattedTextSectionTemplate);
			archiveSection.setLock(true,false);
			FormData formDataArchive = archiveSection.getFormData();
			FormField formField = formDataArchive.get(lang,"st_formatted_text");
		
			Calendar tomorrow = Calendar.getInstance();
			tomorrow = today; // just in case this script is run over midnight...
			tomorrow.add(Calendar.DAY_OF_MONTH,1);
		
			StringBuilder messageBuilder = new StringBuilder();
			messageBuilder.append("<p>").append(THIS_IS_OLD).append("</p>");
			messageBuilder.append("<p><br>Date published: " + prettyDateFormat.format(pageFormData.get(lang,"pt_publish_date").get())).append("</br></p>");
			messageBuilder.append("<p><b>Date superseded: " + prettyDateFormat.format(tomorrow.getTime())).append("</b></p>");
		
			formField.set(messageBuilder.toString());
			archiveSection.setFormData(formDataArchive);
			archiveSection.save();
			archiveSection.setLock(false,false);
			sectionContent.moveChild(archiveSection,sectionContent.getChildIndex(sectionContent.getFirstChild()));
		} catch (Exception sectionex) {
			showMessage(sectionex.getMessage());
			context.logError("Exception while adding date information to page archive.");
			context.logError(sectionex.getMessage());
		}
	}
	*/
	
	page.save("Set this page to be archived",true); 
	page.setLock(false,false);
}

// create new site store folder (navigation item)
PageRefFolder createSiteStoreFolder(IDProvider folder, String name) {
	PageRefFolder newSiteStoreFolder;
	try {
		folder.setLock(true,false);
		newSiteStoreFolder = folder.createPageRefFolder(name,true);
		newSiteStoreFolder.setLock(true,false);
		newSiteStoreFolder.setDisplayName(lang,name);
		newSiteStoreFolder.save();
		newSiteStoreFolder.setLock(false,false);
		folder.moveChild(newSiteStoreFolder,1);		// latest item should appear first in menu
		folder.setLock(false,false);
	} catch (Exception ex) {
		String error = "Exception while creating site store folder inside folder " + folder.getUid() + "\n Exception is : " + ex.getMessage();
		showMessage(error);
	}
	return newSiteStoreFolder;
}

// create the navigation entry pageref
PageRef createPageRef(PageRefFolder refFolder, Page page, String name) {
	PageRef newPageRef;
	try {
		newPageRef = refFolder.createPageRef(refFolder.getUid(),page, true);
		newPageRef.setLock(true,false);
		newPageRef.setDisplayName(lang,name);
		newPageRef.save();
		newPageRef.setLock(false,false);
	} catch (Exception ex) {
		String error = "Exception while creating page reference for page " + page.getUid() + " in folder " + refFolder.getUid() + "\n Exception is : " + ex.getMessage();
		showMessage(error);
	}
	return newPageRef;
}


// get the primary PageRef for the given page. This code is just a guess - what if there are multiple incoming references?
PageRef getPageRef(Page page) {
	PageRef ref = null;
	if(currentPage.hasIncomingReferences()) {
		ReferenceEntry[] incomingRefs = currentPage.getIncomingReferences();
		for(int i = 0; i < incomingRefs.length; i++) {
			if(incomingRefs[i].getType() == ReferenceEntry.SITE_STORE_REFERENCE) {
				ref = incomingRefs[i].getReferencedElement();
				break;
			}
		}
	}
	return ref;
}

// create subheading and link to archive destination page on the original (source) page
createPreviousVersionSectionLinks(Page page, PageRef archiveDestination) {
	try {
		page.setLock(true,false);
		Body sectionContent = page.getBodyByName("pt_section_content");
		SectionTemplate subHeadingST = templateStore.getStoreElement("st_sub_heading",IDProvider.UidType.TEMPLATESTORE);
		SectionTemplate formattedTextSectionTemplate = templateStore.getStoreElement("st_formatted_text",IDProvider.UidType.TEMPLATESTORE);
		
		// set the H3 heading
		Section prevHeading = sectionContent.createSection("previous_versions_heading",subHeadingST);
		prevHeading.setLock(true,false);
		FormData prevHeadingFormData = prevHeading.getFormData();
		FormField prevHeadingFField = prevHeadingFormData.get(lang,"st_sub_heading_text");
		FormField prevHeadingSizeFField = prevHeadingFormData.get(lang,"st_sub_heading_text_size"); // this is a CMS_INPUT_COMBOBOX
		prevHeadingFField.set(HEADING_PREV_VERSIONS);
		prevHeadingSizeFField.set("3");
		prevHeading.setFormData(prevHeadingFormData);
		prevHeading.save();
		prevHeading.setLock(false,false);

		// create the link to previous versions paragraph
		Section prevLinkSection = sectionContent.createSection("previous_versions_link",formattedTextSectionTemplate);
		prevLinkSection.setLock(true,false);
		FormData prevLinkSectionFormData = prevLinkSection.getFormData();
		FormField prevTextField = prevLinkSectionFormData.get(lang,"st_formatted_text");
		prev_lt_internal_link = templateStore.getStoreElement("lt_internal",IDProvider.UidType.TEMPLATESTORE_LINKTEMPLATE);
		prevLink = prev_lt_internal_link.createLink(lang);
		prevLinkFormData = prevLink.getFormData();
		prev_lt_link_type = prevLinkFormData.get(lang,"lt_link_type");
		prev_lt_link_type.set(true);
		prev_lt_title = prevLinkFormData.get(lang,"lt_title");
		prev_lt_title.set(LINK_PREV_VERSIONS);
		prev_lt_text = prevLinkFormData.get(lang,"lt_text");
		prev_lt_text.set(LINK_PREV_VERSIONS);
		prev_lt_reference = prevLinkFormData.get(lang,"lt_reference");
		// this is how you set an FS_REFERENCE value
		archiveRef = TargetReference.TargetReferences.newInstance(lang, archiveDestination, null);
		prev_lt_reference.set(archiveRef);
		prevLink.setFormData(prevLinkFormData);
		// how to add a link to a DOM element?
		dom = prevTextField.get();
		domRoot = dom.getRoot();
		paragraphTemplate = templateStore.getFormatTemplates().getFormatTemplate("p");
		para = domRoot.appendChild(paragraphTemplate);
		para.appendChild(prevLink);
		prevTextField.set(dom);
		
		prevLinkSection.setFormData(prevLinkSectionFormData);
		prevLinkSection.save();
		prevLinkSection.setLock(false,false);
		
		// create "For recent changes to this guidance, please see the bottom of the page."
		Section seeBottom = sectionContent.createSection("sc_see_bottom",formattedTextSectionTemplate);
		seeBottom.setLock(true,false);
		FormData seeBottomFormData = seeBottom.getFormData();
		FormField seeBottomTextField = seeBottomFormData.get(lang,"st_formatted_text");
		seeDom = seeBottomTextField.get();
		seeDomRoot = seeDom.getRoot();
		paragraphTemplate = templateStore.getFormatTemplates().getFormatTemplate("p");
		seePara = seeDomRoot.appendChild(paragraphTemplate);
		seePara.appendChild("For recent changes to this guidance, please see the ");
		
		// create the link to the section on this page
		bottom_lt_internal_link = templateStore.getStoreElement("lt_internal",IDProvider.UidType.TEMPLATESTORE_LINKTEMPLATE);
		bottomLink = bottom_lt_internal_link.createLink(lang);
		bottomLinkFormData = bottomLink.getFormData();
		bottomLink_type = bottomLinkFormData.get(lang,"lt_link_type");
		bottomLink_type.set(true);
		bottomLink_title = bottomLinkFormData.get(lang,"lt_title");
		bottomLink_title.set("See the bottom of the page");
		bottomLink_text = bottomLinkFormData.get(lang,"lt_text");
		bottomLink_text.set("bottom of the page");
		bottomLink_ref = bottomLinkFormData.get(lang,"lt_reference");
	
		
		thisPageRef_link = TargetReference.TargetReferences.newInstance(lang, getPageRef(page), null);
		thisPageRef_link.setSection(prevHeading);
		bottomLink_ref.set(thisPageRef_link);
		bottomLink.setFormData(bottomLinkFormData);
		
		seePara.appendChild(bottomLink);
		
		
		seePara.appendChild(".");
		seeBottomTextField.set(seeDom);
		seeBottom.setFormData(seeBottomFormData);
		seeBottom.save();
		seeBottom.setLock(false,false);
		// and move it to the top of the page
		sectionContent.moveChild(seeBottom,sectionContent.getChildIndex(sectionContent.getFirstChild()));
		page.setLock(false,false);
	} catch (Exception prevex) {
		String error = "Exception while creating previous sections links for " + page.getUid() + " to archive page " + archiveDestination.getUid() + ".\n Exception is : " + prevex.getMessage();
		showMessage(error);
		context.logError(error);
	}
}

/* ************************ Script starts here ***************** */
try {
	if(destination.get() == null || destination.get().get() == null) {
		showMessage("Please select an archive destination folder before creating an archived copy of this page.");
		return;
	}
	currentPage.setLock(true,true);
	
	// format today's date
	String todayString = dateFormat.format(today.getTime());
	
	String archivedName = currentPage.getDisplayName(lang) + " archived " + todayString;
	PageRef destinationPage = destination.get().get().findStartNode();
	
	// determine if there are any existing archive pages by checking for a hidden FS_REFERENCE field
	boolean firstArchive = false;
	FormData pageFormData = currentPage.getFormData();
	FormField prevArchiveRefField = pageFormData.get(lang,"sc_archived_version_reference");
	FormField pubDate = pageFormData.get(lang,"pt_publish_date");
	if(prevArchiveRefField.get() == null || prevArchiveRefField.get().get() == null) {
		firstArchive = true;
	}
	
	PageFolder newPageFolder = createPageStoreFolder(destinationPage.getPage().getParent().getUid(),archivedName);
	if(newPageFolder != null) {
		Page newPage = copyCurrentPage(newPageFolder, currentPage, archivedName);
		if(newPage != null) {
			// assumes only one incoming reference, which should be true of all pages created in ContentCreator
			addArchivedInformationSections(newPage,today,currentPage.getIncomingReferences()[0].getReferencedElement());
			
			PageRefFolder siteStoreFolder = createSiteStoreFolder(destination.get().get(), archivedName);
			if(siteStoreFolder != null) {
				
				if(firstArchive) {
					createPreviousVersionSectionLinks(currentPage,destinationPage);
				}
				currentPage.setLock(true,false);
				
				PageRef newPageRef = createPageRef(siteStoreFolder,newPage,archivedName);
				archiveVersionRef = TargetReference.TargetReferences.newInstance(lang, newPageRef, null);
				prevArchiveRefField.set(archiveVersionRef);
				// update publication date of source page
				Calendar tomorrow = Calendar.getInstance();
				tomorrow = today; // just in case this script is run over midnight...
				tomorrow.add(Calendar.DAY_OF_MONTH,1);
				pubDate.set(tomorrow.getTime());
				
				currentPage.setFormData(pageFormData);
				currentPage.save("Created archived copy of this page with ID " + newPageRef.getUid());
				currentPage.setLock(false,false);
				
				showMessage("Created archived page called: " + newPageRef.getDisplayName(lang));
			} else {
				// all gone wrong;
				return;
			}
		} else {
			// all gone wrong
			return;
		}
	} else {
		// all gone wrong
		return;
	}

} catch (Exception ex) {
	showMessage(ex.getMessage());
	context.logError("Exception while creating page archive.");
	context.logError(ex.getMessage());
}